class Updatable:
    """
    A mixin that can be added to a declared class to add update methods, allowing attributes to be set from
    a :class:`dict`.
    """

    def update(self, **kwargs):
        """Set attributes from the ``kwargs``, ignoring non-existant key/columns."""
        for key, value in kwargs.items():
            if value is DoNotUpdate:
                continue
            if hasattr(self, key):
                setattr(self, key, value)

    def set(self, **kwargs):
        """Set attributes from the ``kwargs``, without checking for non-existant key/columns."""
        for key, value in kwargs.items():
            if value is DoNotUpdate:
                continue
            setattr(self, key, value)


class DoNotUpdateType:
    """
    A type, similar to :data:`None`, used to mark fields that should be skipped in update and set operations.
    """
    __slots__ = ()


DoNotUpdate = DoNotUpdateType()
"""The constant instance of the DoNotUpdateType."""


__all__ = (
    "Updatable",
    "DoNotUpdateType",
    "DoNotUpdate",
)



# Module docstring
"""
Conversations are wrapper classes that can be applied to functions which await
:class:`~royalnet.engineer.bullet.Bullet`\\ s from a :class:`~royalnet.engineer.sentry.Sentry`.
"""

# Special imports
from __future__ import annotations
import royalnet.royaltyping as t

# External imports
import logging

# Internal imports
from . import sentry as s


# Special global objects
log = logging.getLogger(__name__)


# Code
class ConversationProtocol(t.Protocol):
    """
    Typing annotation for Conversation functions.
    """
    def __call__(self, *, _sentry: s.Sentry, **kwargs) -> t.Awaitable[t.Optional[ConversationProtocol]]:
        ...


class Conversation:
    """
    The base class for Conversations. It does nothing on its own except providing better debug information.
    """

    def __init__(self, f: ConversationProtocol, *_, **__):
        self.f: ConversationProtocol = f
        """
        The function that was wrapped by this conversation.
         
        It can be called by calling this object as it was a function::
        
            my_conv(_sentry=_sentry, _msg=msg)
        """

    @classmethod
    def new(cls, *args, **kwargs):
        """
        A decorator that instantiates a new :class:`Conversation` object using the decorated function.

        :return: The created :class:`Conversation` object.
                 It can still be called in the same way as the previous function!
        """
        def decorator(f: ConversationProtocol):
            c = cls(f=f, *args, **kwargs)
            log.debug(f"Created: {c!r}")
            return c
        return decorator

    def __call__(self, *, _sentry: s.Sentry, **kwargs) -> t.Awaitable[t.Optional[ConversationProtocol]]:
        log.debug(f"Calling: {self!r}")
        return self.run(_sentry=_sentry, **kwargs)

    async def run(self, *, _sentry: s.Sentry, **kwargs) -> t.Optional[ConversationProtocol]:
        """
        The coroutine function that generates the coroutines returned by :meth:`.__call__` .

        It is a :class:`Conversation` itself.
        """
        log.debug(f"Running: {self!r}")
        return await self.f(_sentry=_sentry, **kwargs)

    def __repr__(self):
        return f"<{self.__class__.__qualname__} #{id(self)}>"


__all__ = (
    "ConversationProtocol",
    "Conversation"
)



"""
Wrenches are objects which can used instead of coroutines in Sentry receiver filters,
acting similarly to function factories and allowing to easily define filter functions with parameters.
"""

from __future__ import annotations
import royalnet.royaltyping as t

import abc

from . import discard
from . import exc


class Wrench(metaclass=abc.ABCMeta):
    """
    The abstract base class for Wrenches.
    """

    @abc.abstractmethod
    async def filter(self, obj: t.Any) -> t.Any:
        """
        The function applied to all objects transiting through the pipeline:

        - If the function **returns**, its return value will be passed to the next node in the pipeline;
        - If the function **raises**, the error is propagated downwards.

        A special exception is available for discarding objects: :exc:`.discard.Discard`.
        If raised, the object will be silently ignored.
        """
        raise NotImplementedError()

    def __call__(self, obj: t.Any) -> t.Awaitable[t.Any]:
        """
        Allow instances to be directly called, emulating coroutine functions.
        """
        return self.filter(obj)


class PassAll(Wrench):
    """
    **Return** each received object as it is.

    .. note:: To be used only in testing.
    """

    async def filter(self, obj: t.Any) -> t.Any:
        return obj


class DiscardAll(Wrench):
    """
    **Discard** each received object.

    .. note:: To be used only in testing.
    """

    async def filter(self, obj: t.Any) -> t.Any:
        raise discard.Discard(obj, "Discard filter discards everything")


class ErrorAll(Wrench):
    """
    **Raise** :exc:`.exc.DeliberateException` for each received object.

    .. note:: To be used only in testing.
    """

    async def filter(self, obj: t.Any) -> t.Any:
        raise exc.DeliberateException("ErrorAll received an object")


class CheckBase(Wrench, metaclass=abc.ABCMeta):
    """
    Check a condition on the received objects:

    - If the check returns :data:`True`, the object is **returned**;
    - If the check returns :data:`False`, the object is **discarded**;
    - If an error is raised, it is **propagated**.
    """

    def __init__(self, *, invert: bool = False):
        self.invert: bool = invert
        """
        If set to :data:`True`, this Nut will invert its results:
        
        - If the check returns :data:`True`, the object is **discarded**;
        - If the check returns :data:`False`, the object is **returned**;
        - If an error is raised, it is **propagated**.
        """

    def __invert__(self):
        return self.__class__(invert=not self.invert)

    @abc.abstractmethod
    async def check(self, obj: t.Any) -> bool:
        """
        The condition to check.

        :param obj: The object passing through the pipeline.
        :return: Whether the check was successful or not.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def error(self, obj: t.Any) -> str:
        """
        The error message to attach as :attr:`.Discard.message` if the object is discarded.

        :param obj: The object passing through the pipeline.
        :return: The error message.
        """
        raise NotImplementedError()

    async def filter(self, obj: t.Any) -> t.Any:
        if await self.check(obj) ^ self.invert:
            return obj
        else:
            raise discard.Discard(obj=obj, message=self.error(obj))


class Type(CheckBase):
    """
    Check the type of an object:

    - If the object **is** of the specified type, it is **returned**;
    - If the object **isn't** of the specified type, it is **discarded**;
    """

    def __init__(self, type_: t.Type, **kwargs):
        super().__init__(**kwargs)
        self.type: t.Type = type_

    async def check(self, obj: t.Any) -> bool:
        return isinstance(obj, self.type)

    def error(self, obj: t.Any) -> str:
        return f"Not instance of type {self.type}"


class StartsWith(CheckBase):
    """
    Check if an object :func:`startswith` a certain prefix.
    """

    def __init__(self, prefix: str, **kwargs):
        super().__init__(**kwargs)
        self.prefix: str = prefix

    async def check(self, obj: t.Any) -> bool:
        return obj.startswith(self.prefix)

    def error(self, obj: t.Any) -> str:
        return f"Didn't start with {self.prefix}"


class EndsWith(CheckBase):
    """
    Check if an object :func:`endswith` a certain suffix.
    """

    def __init__(self, suffix: str, **kwargs):
        super().__init__(**kwargs)
        self.suffix: str = suffix

    async def check(self, obj: t.Any) -> bool:
        return obj.startswith(self.suffix)

    def error(self, obj: t.Any) -> str:
        return f"Didn't end with {self.suffix}"


class Choice(CheckBase):
    """
    Check if an object is among the accepted list.
    """

    def __init__(self, *accepted, **kwargs):
        super().__init__(**kwargs)
        self.accepted: t.Collection = accepted
        """
        A collection of elements which can be chosen.
        """

    async def check(self, obj: t.Any) -> bool:
        return obj in self.accepted

    def error(self, obj: t.Any) -> str:
        return f"Not a valid choice"


class RegexCheck(CheckBase):
    """
    Check if an object matches a regex pattern.
    """

    def __init__(self, pattern: t.Pattern, **kwargs):
        super().__init__(**kwargs)
        self.pattern: t.Pattern = pattern
        """
        The pattern that should be matched.
        """

    async def check(self, obj: t.Any) -> bool:
        return bool(self.pattern.match(obj))

    def error(self, obj: t.Any) -> str:
        return f"Didn't match pattern {self.pattern}"


class RegexMatch(Wrench):
    """
    Apply a regex over an object:

    - If it matches, **return** the :class:`re.Match` object;
    - If it doesn't match, **discard** the object.
    """

    def __init__(self, pattern: t.Pattern, **kwargs):
        super().__init__(**kwargs)
        self.pattern: t.Pattern = pattern
        """
        The pattern that should be matched.
        """

    async def filter(self, obj: t.Any) -> t.Any:
        if match := self.pattern.match(obj):
            return match
        else:
            raise discard.Discard(obj, f"Didn't match pattern {obj}")


class RegexReplace(Wrench):
    """
    Apply a regex over an object:

    - If it matches, replace the match(es) with :attr:`.replacement` and **return** the result.
    - If it doesn't match, **return** the object as it is.
    """

    def __init__(self, pattern: t.Pattern, replacement: t.Union[str, bytes]):
        self.pattern: t.Pattern = pattern
        """
        The pattern that should be matched.
        """

        self.replacement: t.Union[str, bytes] = replacement
        """
        The substitution string for the object.
        """

    async def filter(self, obj: t.Any) -> t.Any:
        return self.pattern.sub(self.replacement, obj)


class Lambda(Wrench):
    """
    Apply a syncronous function over the received objects.
    """

    def __init__(self, func: t.Callable[[t.Any], t.Any]):
        self.func: t.Callable[[t.Any], t.Any] = func
        """
        The function to apply.
        """

    async def filter(self, obj: t.Any) -> t.Any:
        return self.func(obj)


class Check(CheckBase):
    """
    Check a condition on the received objects.
    """

    def __init__(self, func: t.Callable[[t.Any], t.Any], error: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.func: t.Callable[[t.Any], t.Any] = func
        """
        The condition to check.
        """

        self.error: str = error
        """
        The error message to display if the check fails.
        """

    async def check(self, obj: t.Any) -> bool:
        return self.func(obj)

    def error(self, obj: t.Any) -> str:
        return self.error


__all__ = (
    "Wrench",
    "CheckBase",
    "Type",
    "StartsWith",
    "EndsWith",
    "Choice",
    "RegexCheck",
    "RegexMatch",
    "RegexReplace",
    "Lambda",
    "Check",
)



# Module docstring
"""

"""

# Special imports
from __future__ import annotations
import royalnet.royaltyping as t

# External imports
import logging
import re

# Internal imports
from . import conversation as c
from . import sentry as s
from . import bullet as b
from . import teleporter as tp

# Special global objects
log = logging.getLogger(__name__)


# Code
class Command(c.Conversation):
    """
    A Command is a :class:`~.conversation.Conversation` which is started by a :class:`~.bullet.Message` having a text
    matching the :attr:`.pattern`; the named capture groups of the pattern are then passed as keyword arguments to
    :attr:`.f`.

    .. info:: Usually you don't directly instantiate commands, you define :class:`.PartialCommand`\\ s in packs which
              are later completed by a runner.
    """

    def __init__(self, f: c.ConversationProtocol, *, names: t.List[str] = None, pattern: re.Pattern):
        """
        Create a new :class:`.Command` .
        """
        log.debug(f"Teleporting function: {f!r}")
        teleported = tp.Teleporter(f, validate_input=True, validate_output=False)

        log.debug(f"Initializing Conversation with: {teleported!r}")
        super().__init__(teleported)

        self.names: t.List[str] = names if names else []
        """
        The names of the command, as a :class:`list` of :class:`str`.
        
        The first element of the list is the "main" name, and will be displayed in help messages.
        """

        self.pattern: re.Pattern = pattern
        """
        The pattern that should be matched by the command.
        """

    def name(self):
        """
        :return: The main name of the Command.
        """
        return self.names[0]

    def aliases(self):
        """
        :return: The aliases (non-main names) of the Command.
        """
        return self.names[1:]

    def __repr__(self):
        if (nc := len(self.names)) > 1:
            plus = f" + {nc-1} other names"
        else:
            plus = ""
        return f"<{self.__class__.__qualname__}: {self.name()!r}{plus}>"

    async def run(self, *, _sentry: s.Sentry, **base_kwargs) -> t.Optional[c.ConversationProtocol]:
        log.debug(f"Awaiting a bullet...")
        bullet: b.Bullet = await _sentry

        log.debug(f"Received: {bullet!r}")

        log.debug(f"Ensuring a message was received: {bullet!r}")
        if not isinstance(bullet, b.Message):
            log.debug(f"Returning: {bullet!r} is not a message")
            return

        log.debug(f"Getting message text of: {bullet!r}")
        if not (text := await bullet.text()):
            log.debug(f"Returning: {bullet!r} has no text")
            return

        log.debug(f"Searching for pattern: {text!r}")
        if not (match := self.pattern.search(text)):
            log.debug(f"Returning: Couldn't find pattern in {text!r}")
            return

        log.debug(f"Match successful, getting capture groups of: {match!r}")
        message_kwargs: t.Dict[str, str] = match.groupdict()
        
        log.debug(f"Passing args to function: {message_kwargs!r}")
        return await super().run(_sentry=_sentry, _msg=bullet, **base_kwargs, **message_kwargs)

    def help(self) -> t.Optional[str]:
        """
        Get help about this command. This defaults to returning the docstring of :attr:`.f` .

        :return: The help :class:`str` for this command, or :data:`None` if the command has no docstring.
        """
        return self.f.__doc__


class PartialCommand:
    """
    A PartialCommand is a :class:`.Command` having an unknown :attr:`~.Command.name` and :attr:`~.Command.pattern` at
    the moment of creation.

    They can specified later using :meth:`.complete`.
    """
    def __init__(self, f: c.ConversationProtocol, syntax: str):
        """
        Create a new :class:`.PartialCommand` .

        .. seealso:: :meth:`.new`
        """

        self.f: c.ConversationProtocol = f
        """
        The function to pass to :attr:`.c.Conversation.f`.
        """

        self.syntax: str = syntax
        """
        Part of the pattern from where the arguments should be captured.
        """

    @classmethod
    def new(cls, *args, **kwargs):
        """
        A decorator that instantiates a new :class:`Conversation` object using the decorated function.

        :return: The created :class:`Conversation` object.
                 It can still be called in the same way as the previous function!
        """
        def decorator(f: c.ConversationProtocol):
            partial_command = cls(f=f, *args, **kwargs)
            log.debug(f"Created: {partial_command!r}")
            return partial_command

        return decorator

    def complete(self, *, names: t.List[str] = None, pattern: str) -> Command:
        """
        Complete the :class:`.PartialCommand` with names and a pattern, creating a :class:`Command` object.

        :param names: The names of the command. See :attr:`.Command.names` .
        :param pattern: The pattern to add to the PartialCommand. It is first :meth:`~str.format`\\ ted with the keyword
                        arguments ``name`` and ``syntax`` and later :func:`~re.compile`\\ d with the
                        :data:`re.IGNORECASE` flag.
        :return: The complete :class:`Command`.
        """
        if names is None:
            names = []
        if len(names) < 1:
            raise ValueError("Commands must have at least one name.")
        for name in names:
            if not name.isalnum():
                raise ValueError(f"Name is not alphanumeric: {name!r}")

        name_regex = f"(?:{'|'.join(names)})"
        pattern: re.Pattern = re.compile(pattern.format(name=name_regex, syntax=self.syntax), re.IGNORECASE)
        return Command(f=self.f, names=names, pattern=pattern)

    def __repr__(self):
        return f"<{self.__class__.__qualname__} {self.f!r}>"


# Objects exported by this module
__all__ = (
    "Command",
    "PartialCommand",
)



# Module docstring
"""
Magazines are references to the bullet classes used by a specific frontend; they allow bullets to instance each other
without tying the instantiations to specific classes.
"""

# Special imports
from __future__ import annotations
import royalnet.royaltyping as t

# External imports
import functools
import logging

# Internal imports
from . import bullet

# Special global objects
log = logging.getLogger(__name__)


# Code
# noinspection PyPep8Naming
class Magazine:
    """
    A reference to all types of bullets to be used when instancing bullets from a bullet.
    """

    _BULLET = bullet.Bullet
    _USER = bullet.User
    _MESSAGE = bullet.Message
    _CHANNEL = bullet.Channel

    @property
    def Bullet(self) -> functools.partial[bullet.Bullet]:
        log.debug(f"Extracting Bullet from Magazine: {self._BULLET!r}")
        return functools.partial(self._BULLET, mag=self)

    @property
    def User(self) -> functools.partial[bullet.User]:
        log.debug(f"Extracting User from Magazine: {self._USER!r}")
        return functools.partial(self._USER, mag=self)

    @property
    def Message(self) -> functools.partial[bullet.Message]:
        log.debug(f"Extracting Message from Magazine: {self._MESSAGE!r}")
        return functools.partial(self._MESSAGE, mag=self)

    @property
    def Channel(self) -> functools.partial[bullet.Channel]:
        log.debug(f"Extracting Channel from Magazine: {self._CHANNEL!r}")
        return functools.partial(self._CHANNEL, mag=self)


# Objects exported by this module
__all__ = (
    "Magazine",
)



"""
This module adds some new common royaltyping to the default typing package.

It should be imported with: ::

    from royalnet.royaltyping import *

"""

from typing import *
# noinspection PyUnresolvedReferences
from typing import IO, TextIO, BinaryIO


JSONScalar = Union[
    None,
    float,
    int,
    str,
]
"""
A non-recursive JSON value: either :data:`None`, a :class:`float`, a :class:`int` or a :class:`str`.
"""

JSON = Union[
    JSONScalar,
    List["JSON"],
    Dict[str, "JSON"],
]
"""
A recursive JSON value: either a :data:`.JSONScalar`, or a :class:`list` of :data:`.JSON` objects, or a :class:`dict` 
of :class:`str` to :data:`.JSON` mappings. 
"""



"""
This module defines multiple objects that simplify some :mod:`sqlalchemy` tasks.
"""

from .func import *
from .make import *
from .repr import *
from .update import *

__all__ = (
    "ieq",
    "ineq",
    "Makeable",
    "ColRepr",
    "Updatable",
    "DoNotUpdateType",
    "DoNotUpdate",
)



"""
Sentries are asynchronous receivers for events (usually :class:`bullet.Bullet`) incoming from Dispensers.

They support event filtering through Wrenches and coroutine functions.
"""

from __future__ import annotations
import royalnet.royaltyping as t

import abc
import logging
import asyncio

from . import discard

if t.TYPE_CHECKING:
    from .dispenser import Dispenser

log = logging.getLogger(__name__)


class Sentry(metaclass=abc.ABCMeta):
    """
    The abstract object representing a node of the pipeline.
    """

    @abc.abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_nowait(self):
        """
        Try to get a single :class:`~.bullet.Bullet` from the pipeline, without blocking or handling discards.

        :return: The **returned** :class:`~.bullet.Bullet`.
        :raises asyncio.QueueEmpty: If the queue is empty.
        :raises .discard.Discard: If the object was **discarded** by the pipeline.
        :raises Exception: If an exception was **raised** in the pipeline.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    async def get(self):
        """
        Try to get a single :class:`~.bullet.Bullet` from the pipeline, blocking until something is available, but
        without handling discards.

        :return: The **returned** :class:`~.bullet.Bullet`.
        :raises .discard.Discard: If the object was **discarded** by the pipeline.
        :raises Exception: If an exception was **raised** in the pipeline.
        """
        raise NotImplementedError()

    async def wait(self):
        """
        Try to get a single :class:`~.bullet.Bullet` from the pipeline, blocking until something is available and is not
        discarded.

        :return: The **returned** :class:`~.bullet.Bullet`.
        :raises Exception: If an exception was **raised** in the pipeline.
        """
        while True:
            try:
                result = await self.get()
                log.debug(f"Returned: {result}")
                return result
            except discard.Discard as d:
                log.debug(f"{str(d)}")
                continue

    def __await__(self):
        """
        Awaiting an object implementing :class:`.SentryInterface` corresponds to awaiting :meth:`.wait`.
        """
        return self.get().__await__()

    @abc.abstractmethod
    async def put(self, item: t.Any) -> None:
        """
        Insert a new item in the queue.

        :param item: The item to be added.
        """
        raise NotImplementedError()

    def filter(self, wrench: t.Callable[[t.Any], t.Awaitable[t.Any]]) -> SentryFilter:
        """
        Chain a new filter to the pipeline.

        :param wrench: The filter to add to the chain. It can either be a :class:`.wrench.Wrench`, or a coroutine
                       function accepting a single object as parameter and returning the same or a different one.
        :return: A new :class:`.SentryFilter` which includes the filter.

        .. seealso:: :meth:`.__or__`
        """
        if callable(wrench):
            return SentryFilter(previous=self, wrench=wrench)
        else:
            raise TypeError("wrench must be either a Wrench or a coroutine function")

    def __or__(self, other: t.Callable[[t.Any], t.Awaitable[t.Any]]) -> SentryFilter:
        """
        A unix-pipe-like interface for :meth:`.filter`.

        .. code-block::

           await (sentry | wrench.Type(Message) | wrench.Sync(lambda o: o.text))

        """
        try:
            return self.filter(other)
        except TypeError:
            raise TypeError("Right-side must be either a Wrench or a coroutine function")

    @abc.abstractmethod
    def dispenser(self):
        """
        Get the :class:`.Dispenser` that created this Sentry.

        :return: The :class:`.Dispenser` object.
        """
        raise NotImplementedError()


class SentryFilter(Sentry):
    """
    A non-root node of the filtering pipeline.
    """

    def __init__(self, previous: Sentry, wrench: t.Callable[[t.Any], t.Awaitable[t.Any]]):
        self.previous: Sentry = previous
        """
        The previous node of the pipeline.
        """

        self.wrench: t.Callable[[t.Any], t.Awaitable[t.Any]] = wrench
        """
        The coroutine function to apply to all objects passing through this node.
        """

    def __len__(self) -> int:
        return len(self.previous) + 1

    def get_nowait(self):
        return self.previous.get_nowait()

    async def get(self):
        return await self.previous.get()

    async def put(self, item) -> None:
        return await self.previous.put(item)

    def dispenser(self):
        return self.previous.dispenser()


class SentrySource(Sentry):
    """
    The root and source of the pipeline.
    """

    def __init__(self, dispenser: "Dispenser", queue_size: int = 12):
        self.queue: asyncio.Queue = asyncio.Queue(maxsize=queue_size)
        self._dispenser: "Dispenser" = dispenser

    def __len__(self) -> int:
        return 1

    def get_nowait(self):
        return self.queue.get_nowait()

    async def get(self):
        return await self.queue.get()

    async def put(self, item) -> None:
        return await self.queue.put(item)

    async def dispenser(self):
        return self._dispenser


__all__ = (
    "Sentry",
    "SentryFilter",
    "SentrySource",
)



"""
Dispensers instantiate sentries and dispatch events in bulk to the whole group.
"""

from __future__ import annotations
import royalnet.royaltyping as t

import logging
import contextlib

from .sentry import SentrySource
from .conversation import Conversation

log = logging.getLogger(__name__)


class Dispenser:
    def __init__(self):
        self.sentries: t.List[SentrySource] = []
        """
        A :class:`list` of all the running sentries of this dispenser.
        """

    async def put(self, item: t.Any) -> None:
        """
        Insert a new item in the queues of all the running sentries.

        :param item: The item to insert.
        """
        log.debug(f"Putting {item}...")
        for sentry in self.sentries:
            await sentry.put(item)

    @contextlib.contextmanager
    def sentry(self, *args, **kwargs):
        """
        A context manager which creates a :class:`.SentrySource` and keeps it in :attr:`.sentries` while it is being
        used.
        """
        log.debug("Creating a new SentrySource...")
        sentry = SentrySource(dispenser=self, *args, **kwargs)

        log.debug(f"Adding: {sentry}")
        self.sentries.append(sentry)

        log.debug(f"Yielding: {sentry}")
        yield sentry

        log.debug(f"Removing from the sentries list: {sentry}")
        self.sentries.remove(sentry)

    async def run(self, conv: Conversation, **kwargs) -> None:
        """
        Run the passed conversation.

        :param conv: The conversation to run.
        """
        log.debug(f"Running: {conv}")
        with self.sentry() as sentry:
            state = conv(_sentry=sentry, **kwargs)

            log.debug(f"First state: {state}")
            while state := await state:
                log.debug(f"Switched to: {state}")


__all__ = (
    "Dispenser",
)



"""
Bullets are parts of the data model that :mod:`royalnet.engineer` uses to build a common interface between different
chat apps (*frontends*).

They exclusively use coroutine functions to access data, as it may be required to fetch it from a remote location before
it is available.

**All** coroutine functions can have three different results:

- :exc:`.exc.BulletException` is raised, meaning that something went wrong during the data retrieval.
  - :exc:`.exc.NotSupportedError` is raised, meaning that the frontend does not support the feature the requested data
    is about (asking for :meth:`.Message.reply_to` in an IRC frontend, for example).
- :data:`None` is returned, meaning that there is no data in that field (if a message is not a reply to anything,
  :meth:`Message.reply_to` will be :data:`None`.
- The data is returned.

To instantiate a new :class:`Bullet` from a bullet, you should use the methods of :attr:`.Bullet.mag`.
"""

from __future__ import annotations
import royalnet.royaltyping as t

import abc
import datetime
import sqlalchemy.orm

from . import exc

if t.TYPE_CHECKING:
    from . import magazine


class Bullet(metaclass=abc.ABCMeta):
    """
    The abstract base class for :mod:`~royalnet.engineer.bullet` models.
    """

    def __init__(self, mag: "magazine.Magazine"):
        """
        Instantiate a new :class:`.Bullet` .
        """

        self.mag: "magazine.Magazine" = mag
        """
        The :class:`.magazine.Magazine` to use when instantiating new :class:`.Bullet`\\ s.
        """

    @abc.abstractmethod
    def __hash__(self) -> int:
        """
        :return: A value that uniquely identifies the object in this Python interpreter process.
        """
        raise NotImplementedError()


class Message(Bullet, metaclass=abc.ABCMeta):
    """
    An abstract class representing a chat message.
    """

    async def text(self) -> t.Optional[str]:
        """
        :return: The raw text contents of the message.
        """
        raise exc.NotSupportedError()

    async def timestamp(self) -> t.Optional[datetime.datetime]:
        """
        :return: The :class:`datetime.datetime` at which the message was sent.
        """
        raise exc.NotSupportedError()

    async def reply_to(self) -> t.Optional[Message]:
        """
        :return: The :class:`.Message` this message is a reply to.
        """
        raise exc.NotSupportedError()

    async def channel(self) -> t.Optional[Channel]:
        """
        :return: The :class:`.Channel` this message was sent in.
        """
        raise exc.NotSupportedError()

    async def files(self) -> t.Optional[t.List[t.BinaryIO]]:
        """
        :return: A :class:`list` of files attached to the message.
        """
        raise exc.NotSupportedError()

    async def send_reply(self, *,
                         text: str = None,
                         files: t.List[t.BinaryIO] = None) -> t.Optional[Message]:
        """
        Reply to this message in the same channel it was sent in.

        :param text: The text to reply with.
        :param files: A :class:`list` of files to attach to the message. The file type should be detected automatically
                      by the frontend, and sent in the best format possible (if all files are photos, they should be
                      sent as a photo album, etc.).
        :return: The sent reply message.
        """
        raise exc.NotSupportedError()


class Channel(Bullet, metaclass=abc.ABCMeta):
    """
    An abstract class representing a channel where messages can be sent.
    """

    async def name(self) -> t.Optional[str]:
        """
        :return: The name of the message channel, such as the chat title.
        """
        raise exc.NotSupportedError()

    async def topic(self) -> t.Optional[str]:
        """
        :return: The topic (description) of the message channel.
        """
        raise exc.NotSupportedError()

    async def users(self) -> t.List[User]:
        """
        :return: A :class:`list` of :class:`.User` who can read messages sent in the channel.
        """
        raise exc.NotSupportedError()

    async def send_message(self, *,
                           text: str = None,
                           files: t.List[t.BinaryIO] = None) -> t.Optional[Message]:
        """
        Send a message in the channel.

        :param text: The text to send in the message.
        :param files: A :class:`list` of files to attach to the message. The file type should be detected automatically
                      by the frontend, and sent in the best format possible (if all files are photos, they should be
                      sent as a photo album, etc.).
        :return: The sent message.
        """
        raise exc.NotSupportedError()


class User(Bullet, metaclass=abc.ABCMeta):
    """
    An abstract class representing a user who can read or send messages in the chat.
    """

    async def name(self) -> t.Optional[str]:
        """
        :return: The user's name.
        """
        raise exc.NotSupportedError()

    async def database(self, session: sqlalchemy.orm.Session) -> t.Any:
        """
        :param session: A :class:`sqlalchemy.orm.Session` instance to use to fetch the database entry.
        :return: The database entry for this user.
        """
        raise exc.NotSupportedError()

    async def send_message(self, *,
                           text: str = None,
                           files: t.List[t.BinaryIO] = None) -> t.Optional[Message]:
        """
        Send a private message to the user.

        :param text: The text to send in the message.
        :param files: A :class:`list` of files to attach to the message. The file type should be detected automatically
                      by the frontend, and sent in the best format possible (if all files are photos, they should be
                      sent as a photo album, etc.).
        :return: The sent message.
        """
        raise exc.NotSupportedError()


__all__ = (
    "Bullet",
    "Message",
    "Channel",
    "User",
)



from royalnet.royaltyping import *
import sqlalchemy.orm as o


T = TypeVar('T')


class Makeable:
    """
    A mixin that can be added to a declared class to add the make and unmake methods, that try find an item with
    specific properties and either create it if it doesn't exist or delete it if it exists.
    """

    @classmethod
    def make(cls: Type[T], session: o.session.Session, **kwargs) -> T:
        """
        Find the item with the specified name, or create it if it doesn't exist.

        :param session: The session to be used in the query and creation.
        :param kwargs: Arguments to use in the :meth:`~sqlalchemy.orm.query.Query.filter_by` clause and in the item
                       constructor.
        :return: The retrieved or created item.
        """
        # Find the item
        item = session.query(cls).filter_by(**kwargs).one_or_none()
        # Create the item
        if item is None:
            item = cls(**kwargs)
            session.add(item)
        # Return the item
        return item

    @classmethod
    def unmake(cls: Type[T], session: o.session.Session, **kwargs) -> None:
        """
        Find the item with the specified name, and delete it if it exists.

        :param session: The session to be used in the query and creation.
        :param kwargs: Arguments to use in the :meth:`~sqlalchemy.orm.query.Query.filter_by` clause and in the item
                       constructor.
        """
        # Find the item
        item = session.query(cls).filter_by(**kwargs).one_or_none()
        # Delete the item
        if item is not None:
            session.delete(item)


__all__ = (
    "Makeable",
)
